import logging
from typing import List
from openlineage.airflow.extractors.base import TaskMetadata
from alvin_integration.producers.airflow.lineage.extractors.mixins import AlvinAirflowExtractorMixin
from alvin_integration.models import AlvinFacet, AlvinBigQueryLineageDetails
from openlineage.airflow.extractors.bigquery_extractor import BigQueryExtractor
from openlineage.airflow.utils import get_job_name
from openlineage.client.facet import SqlJobFacet
from openlineage.common.provider.bigquery import (
    BigQueryDatasetsProvider,
    BigQueryErrorRunFacet,
    BigQueryFacets,
)

log = logging.getLogger(__name__)


class AlvinBigQueryBaseExtractor(AlvinAirflowExtractorMixin, BigQueryExtractor):
    """
    This class contains common logic to handle extraction of BigQuery lineage
    for both Airflow 1 and 2.

    https://cloud.google.com/bigquery/docs/information-schema-jobs
    """

    def __init__(self, operator, *args, **kwargs):
        super().__init__(operator, *args, **kwargs)

    @classmethod
    def get_operator_classnames(cls) -> List[str]:
        raise NotImplementedError

    def build_task_metadata(self, task_instance):
        if not self.operator.hook:
            return TaskMetadata(
                name=get_job_name(task=self.operator),
                inputs=[],
                outputs=[],
                run_facets={}
            )
        bigquery_job_id = self._get_xcom_bigquery_job_id(task_instance)
        context = self.parse_sql_context()
        big_query_client = self.operator.hook.get_client(
                project_id=self.operator.hook.project_id
            )

        stats = BigQueryDatasetsProvider(big_query_client).get_facets(bigquery_job_id)

        return TaskMetadata(
                    name=get_job_name(task=self.operator),
                    inputs=[ds.to_openlineage_dataset() for ds in stats.inputs],
                    outputs=[stats.output.to_openlineage_dataset()] if stats.output else [],
                    run_facets=stats.run_facets,
                    job_facets={'sql': SqlJobFacet(context.sql)}
                )

    def build_facet(self, task_instance):
        return AlvinFacet(
            alvin=AlvinBigQueryLineageDetails(
                job_id=self._get_xcom_bigquery_job_id(task_instance),
                project_id=self.operator.hook.project_id if self.operator.hook else None,
                execution=self.get_execution_details(task_instance),
                connection_id=self.get_connection_id()
            )
        )


class AlvinBigQueryExtractor(AlvinBigQueryBaseExtractor):
    """
    This class contains logic for Airflow 2.

    Note that on this version, the BigQuery operator class is
    part of this package: apache-airflow-providers-google.

    https://github.com/OpenLineage/OpenLineage/tree/main/integration/airflow#custom-extractors

    The following env var needs to be set in the airflow host environment:
    OPENLINEAGE_EXTRACTOR_BigQueryExecuteQueryOperator
        =alvin_airflow.lineage.extractors.bigquery.AlvinBigQueryExtractor
    """
    def get_airflow_run_id(self, task_instance):
        return task_instance.run_id

    def get_connection_id(self):
        return self.operator.gcp_conn_id if self.operator else None

    @classmethod
    def get_operator_classnames(cls) -> List[str]:
        return ["BigQueryExecuteQueryOperator", "BigQueryInsertJobOperator"]


class AlvinBigQueryLegacyExtractor(AlvinBigQueryBaseExtractor):
    """
    This class contains logic for Airflow 1.

    Note that on this version, the BigQuery operator class part of the
    Airflow contrib google module.

    https://github.com/OpenLineage/OpenLineage/tree/main/integration/airflow#custom-extractors

    The following env var needs to be set in the airflow host environment:
    OPENLINEAGE_EXTRACTOR_BigQueryOperator
        =alvin_airflow.lineage.extractors.bigquery.AlvinBigQueryLegacyExtractor
    """
    def get_airflow_run_id(self, task_instance):
        return task_instance.get_dagrun().run_id

    def get_connection_id(self):
        return self.operator.bigquery_conn_id if self.operator else None

    @classmethod
    def get_operator_classnames(cls) -> List[str]:
        return ["BigQueryOperator"]



import requests
import os
from alvin_integration.producers.airflow.config import ALVIN_BACKEND_API_URL, ALVIN_BACKEND_API_KEY


ALVIN_PLATFORM_ID = os.getenv('ALVIN_PLATFORM_ID')


def get_airflow_version():
    import airflow
    return airflow.__version__


def get_dag_source_code(dag_location):
    if dag_location:
        with open(dag_location, 'r') as f:
            content = f.read()
        return content


def extract_dag_metadata(session):
    from airflow.models.serialized_dag import SerializedDagModel
    from airflow.serialization.serialized_objects import SerializedDAG

    dag_metadata = SerializedDagModel.read_all_dags()

    alvin_backend_metadata_url = f'{ALVIN_BACKEND_API_URL}/api/v1/metadata'
    entities = []
    for dag_id, serialized_dag in dag_metadata.items():
        row = SerializedDagModel.get(dag_id, session)
        dag = row.dag
        dag_metadata = SerializedDAG.to_dict(dag)
        dag_metadata['dag']['source_code'] = get_dag_source_code(dag_metadata['dag'].get('fileloc'))
        entities.append(dag_metadata['dag'])
    payload = {'alvin_platform_id': ALVIN_PLATFORM_ID,
               'facets': {'airflow_version': get_airflow_version(),
                          'platform_type': 'AIRFLOW'},
               'entities': entities}
    requests.post(alvin_backend_metadata_url, json=payload, headers={"X-API-KEY": ALVIN_BACKEND_API_KEY})
    print(f'Metadata extracted: {payload}')



from openlineage.lineage_backend import Backend
import logging
import traceback
import pathlib
import os
from airflow.lineage.backend import LineageBackend

log = logging.getLogger(__name__)


class AlvinAirflowBackend(Backend):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class AlvinBackendMeta(type):

    def __init__(cls, *args, **kwargs):
        from airflow.configuration import conf
        backend_str = conf.get("lineage", "backend")
        if backend_str and cls.__name__ in backend_str:
            cls.copy_alvin_plugin()

    @staticmethod
    def read_alvin_plugin():
        """Returns the alvin plugin path."""
        filepath = os.path.join(f'{pathlib.Path(__file__).resolve().parent}', 'alvin_plugin.py')

        return filepath

    def copy_alvin_plugin(self):
        """Copy the Alvin plugin to the airflow plugins folder."""
        filepath = self.read_alvin_plugin()
        plugin_path = os.path.join(f'{os.getenv("AIRFLOW_HOME")}', 'plugins')
        if not os.path.exists(plugin_path):
            log.info(f'Creating plugins folder: {plugin_path}')
            os.makedirs(plugin_path)
        full_path = os.path.join(f'{plugin_path}', 'alvin_plugin.py')
        with open(full_path, 'w') as plugin_file:
            with open(filepath, 'r') as f:
                content = f.read()
                plugin_file.write(content)
        log.info(f'Plugin copied successfully to: {full_path}')


class AlvinBackendComposerMeta(AlvinBackendMeta):

    def copy_alvin_plugin(self):
        from alvin_integration.producers.airflow.config import GOOGLE_COMPOSER_BUCKET
        from google.cloud import storage

        print(f'Creating Alvin DAGs on Google Composer path {GOOGLE_COMPOSER_BUCKET}.....')

        filepath = self.read_alvin_plugin()

        with open(filepath, 'r') as f:
            content = f.read()

            client = storage.Client()

            bucket = client.get_bucket(GOOGLE_COMPOSER_BUCKET)

            blob = bucket.blob('plugins/alvin_plugin.py')

            blob.upload_from_string(content)

        print(f'Plugin copied successfully to: {GOOGLE_COMPOSER_BUCKET}/plugins/alvin_plugin.py')


class AlvinAirflowLineageBackend(LineageBackend, metaclass=AlvinBackendMeta):
    backend: AlvinAirflowBackend = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def send_lineage(cls, *args, **kwargs):
        """Alvin implementation of the send_lineage backend."""
        log.info("calling send lineage with args and kwargs")
        if not cls.backend:
            log.info("Lineage backend is none, creating it")
            cls.backend = AlvinAirflowBackend()
        return cls.backend.send_lineage(*args, **kwargs)


class AlvinAirflowLineageBackendComposer(AlvinAirflowLineageBackend, metaclass=AlvinBackendComposerMeta):
    pass


def alvin_callback(context, operator, is_airflow_legacy=False):
    """Alvin callback implementation.

    This function use the OpenLineage implementation of
    the Airflow Backend and call send_lineage with the
    given context.
    """
    try:
        if is_airflow_legacy:
            AlvinAirflowLineageBackend.send_lineage(context=context, operator=operator)  # noqa
        else:
            lineage_backend = AlvinAirflowLineageBackend()
            lineage_backend.send_lineage(context=context, operator=operator)
    except Exception:
        log.error(f"Alvin Callback Error: {traceback.format_exc()}")


def alvin_dag_run_extractor(dag_run):
    import requests
    import os
    from alvin_integration.producers.airflow.config import ALVIN_BACKEND_API_URL, ALVIN_BACKEND_API_KEY
    ALVIN_PLATFORM_ID = os.getenv('ALVIN_PLATFORM_ID')
    alvin_backend_metadata_url = f'{ALVIN_BACKEND_API_URL}/api/v1/lineage'

    payload = {'alvin_platform_id': ALVIN_PLATFORM_ID,
               'facets': {
                   "dag_run": {
                    "dag_id": dag_run.dag_id,
                    "run_id": dag_run.run_id,
                    "queued_at": (dag_run.queued_at.isoformat() if dag_run.queued_at else None) if hasattr(dag_run, 'queued_at') else None,
                    "execution_date": dag_run.execution_date.isoformat(),
                    "start_date": dag_run.start_date.isoformat(),
                    "end_date": dag_run.end_date.isoformat() if dag_run.end_date else None,
                    "state": dag_run.get_state()
                   }
               }}
    requests.post(alvin_backend_metadata_url, json=payload, headers={"X-API-KEY": ALVIN_BACKEND_API_KEY})
